{

Project: Huffman.dpr
Current file: unit_implement.pas
Author: Poirot
E-Mail: poirot@rol.ru
Copyrights @ 14.05.2002

}

unit unit_implement;

interface

 uses SysUtils, Dialogs;

 type

  PhufElement = ^ThufElement; //  ThufElement
  ThufElement = record        //  
   Letter:Char;               //,    #0   
   Capacity:Longword;         //-         
   NextLeft,                  //  . . (  2-)
   NextRight:PhufElement;      //  . . (  2-)
  end;

  TCharCapHash = array [char] of LongWord; //-  
  TCurUnits    = array of PhufElement;
  TBinarHashRecord = array [0..7] of Byte;
  TBufString = String[8];

  TBinHash =array [char] of TBufString;

 const
  States:array [0..7] of String =
   (
    ' ...',
    ' mutex...',
    '...',
    ' o...',
    ' ...',
    '  - %d...',
    ' ...',
    ' %s ...'
   );
 // 
 procedure Swap(var Value1, Value2: Byte);

 //MinF, MinR   2   Capacity   CurUnits
 procedure Get2MinElement( Len:Word; var Cur:TCurUnits; var MinF, MinS: Byte);

 //  TCharCapHash
 procedure ClearCharCapHash(var CharHash:TCharCapHash);

 //-  0    TCharCapHash
 function GetNonZeroCount(CharHash:TCharCapHash):Word;

 //   TCurUnits  TCharCapHash  0 
 procedure FillCurUnits(var Cur:TCurUnits;CharHash:TCharCapHash);

 // TCurUnits
 procedure ClearCurUnits(var Cur:TCurUnits);

 //  ...  2-  -     
 function BuildTreeFrom2Min(Len:Word; var Cur:TCurUnits):Byte;

 // 
 procedure ClearHuffmanTree(var Head:PhufElement);

 // TBinHash
 procedure ClearBinHash(var Bin:TBinHash);

 //  
 procedure CreateTablOf(var Head:PhufElement; Prev:TBufString; var Bin:TBinHash);

 //   Zero One  
 function GetZeroOneZFileSize(var Bin:TBinHash; var CharHash:TCharCapHash):LongWord;

 //     ()
 function Get01toChar(BinArr:TBinarHashRecord):Byte;

implementation
  // 
  procedure Swap(var Value1, Value2: Byte);
   var Tmp:byte;
   begin
    Tmp:=Value1;
    Value1:=Value2;
    Value2:=Tmp;
   end;
  //MinF, MinR   2   Capacity   CurUnits
  procedure Get2MinElement(Len:Word; var Cur:TCurUnits; var MinF, MinS: Byte);
   var MF,      //   . - CurUnit
       MS,      //   . - CurUnit
       I:Byte;
   begin

    if Cur[0]^.Capacity < Cur[1]^.Capacity then
      begin MF:=0; MS:=1  end
     else
      begin MF:=1; MS:=0  end;

    for I:=2 to Len-1 do
     begin
      if Cur[I]^.Capacity < Cur[MS]^.Capacity then
       begin
        if Cur[I]^.Capacity <= Cur[MF]^.Capacity then
          begin MS:=MF; MF:=I  end
         else MS:=I;
       end;
     end;

    MinF:=MF;
    MinS:=MS;
   end;

  //  TCharCapHash
  procedure ClearCharCapHash(var CharHash:TCharCapHash);
   var i:Char;
   begin
    for i:=#0 to #255 do CharHash[i]:=0;
   end;

  //-  0    TCharCapHash
  function GetNonZeroCount(CharHash:TCharCapHash):Word;
   var I:Char;
       Res:Word;
   begin
    Res:=0;
    for I:=#0 to #255 do
     if CharHash[I]<>0 then Inc(Res);
    Result:=Res;
   end;

  //   TCurUnits  TCharCapHash  0 
  procedure FillCurUnits(var Cur:TCurUnits;CharHash:TCharCapHash);
   var I :Char;
       CHPos:Byte;
       Tmp:PhufElement;
   begin
    SetLength(Cur,GetNonZeroCount(CharHash));
    CHPos:=0;
    for I:=#0 to #255 do
     if CharHash[I]<>0 then
      begin
       New(Tmp);
       with Tmp^ do
        begin
         Capacity:=CharHash[I];
         Letter:=I;
         NextLeft:=nil;
         NextRight:=nil;
        end;
       Cur[CHPos]:=Tmp;
       Inc(CHPos);
      end;
   end;

  // TCurUnits
  procedure ClearCurUnits(var Cur:TCurUnits);
   var I:Byte;
   begin
    for I:=1 to Length(Cur)-1 do Dispose(Cur[I])
   end;

 //  ...  2-  -     
 function BuildTreeFrom2Min(Len:Word; var Cur:TCurUnits):Byte;
  var TmpF,TmpS:PhufElement;
      MinF,MinS,I, AbsIMF, AbsIMS:Byte;

  procedure CopyRecord(var Target, Res:PhufElement);
   begin
    Target^.Capacity:=Res^.Capacity;
    Target^.Letter:=Res^.Letter;
    Target^.NextLeft:=Res^.NextLeft;
    Target^.NextRight:=Res^.NextRight
   end;

  begin
   Get2MinElement(Len,Cur,MinF,MinS);
   New(TmpF);
   New(TmpS);
   CopyRecord(TmpF,Cur[MinF]);
   CopyRecord(TmpS,Cur[MinS]);

   if MinF>MinS then
     begin
      AbsIMF:=MinS;
      AbsIMS:=MinF
     end
    else
     begin
      AbsIMF:=MinF;
      AbsIMS:=MinS
     end;

   with Cur[AbsIMF]^ do
   begin
    Capacity:=TmpF^.Capacity+TmpS^.Capacity;
    NextLeft:=TmpF;
    NextRight:=TmpS;
   end;

   for I:=AbsIMS to (Len-1)-1 do
    CopyRecord(Cur[i],Cur[I+1]);

   Result:=Len-1;
  end;

  // 
  procedure ClearHuffmanTree(var Head:PhufElement);
   begin
    if Head<>nil then
     begin
      ClearHuffmanTree(Head^.NextLeft);
      ClearHuffmanTree(Head^.NextRight);
      Dispose(Head);
     end;
   end;

  // TBinHash
  procedure ClearBinHash(var Bin:TBinHash);
  var I:Char;
  begin
   for I:=#0 to #255 do Bin[I]:='';
  end;

  //  
  procedure CreateTablOf(var Head:PhufElement; Prev:TBufString; var Bin:TBinHash);
   begin
    if Head<>nil then
     begin
      CreateTablOf(Head^.NextLeft,Prev+'0',Bin);
      CreateTablOf(Head^.NextRight,Prev+'1',Bin);
      if Bin[Head^.Letter]='' then Bin[Head^.Letter]:=Prev;
     end;
   end;

 //   Zero One  
 function GetZeroOneZFileSize(var Bin:TBinHash; var CharHash:TCharCapHash):LongWord;
  var
   I:Char;
   Size:LongWord;
  begin
   Size:=0;
   for I:=#0 to #255 do
    if CharHash[I]<>0 then
      Size:=Size+CharHash[I]*Length(Bin[I]);
   Result:=Size;
  end;

 //     ()
 function Get01toChar(BinArr:TBinarHashRecord):Byte;
  var
   I, Sum:Byte;
   Degree:word;
  begin
   Degree:=1;
   Sum:=0;
   for I:=7 downto 0 do
    begin
     Sum:=Sum+BinArr[i]*Degree;
     Degree:=Degree*2;
    end;
    Result:=Sum;
  end;

end.
